theory Explore_TreeT
imports Tutorial_Basic
begin
  text \<open>
    Note that we represent directed graphs as relations of type @{typ "'a\<times>'a"}.
    The successors of a node can then be expressed using the relation-image operator
    @{term "op ``"}, i.e. we have @{lemma "E``{u} = {v . (u,v)\<in>E}" by auto}.
    \<close>

  subsection \<open>Recursion combinator\<close>
  context
    fixes E :: "('a\<times>'a) set" and v0 vt :: 'a
  begin  

  definition "fe_inv v \<equiv> (\<lambda>it r. (\<forall>u\<in>it. (v0,u)\<in>E\<^sup>*) \<and> (r \<longrightarrow> (v,vt)\<in>E\<^sup>*))"
  (* DFS to check path from v0 to vt, with REC, FOREACH. Annotate invar *)
  definition "dfs_abs \<equiv> do {
    REC (\<lambda>dfs v. do {
      if v = vt then
        RETURN True
      else do {
        FOREACHci (fe_inv v)
          ({u. (v,u)\<in>E}) (\<lambda>r. \<not>r) (\<lambda>u r. do {
          dfs u
        }) False
      }  
    }) v0
  }"

  (* Correctness lemma, easy direction. *)
  lemma
    assumes [simp]: "\<And>u. finite {v. (u,v)\<in>E}"
    shows "dfs_abs \<le> SPEC (\<lambda>r. r \<longrightarrow> (v0,vt)\<in>E\<^sup>*)"
    unfolding dfs_abs_def fe_inv_def
    apply (refine_vcg
      REC_rule[where pre="\<lambda>v. (v0,v)\<in>E\<^sup>*"]
      )
    apply auto []
    apply auto []
    apply auto []
    apply auto []
    apply auto []

    apply (rule order_trans)
    apply rprems
    apply auto []
    apply auto []
    thm FOREACHc_rule
    apply auto []
    apply auto []
    done

  definition "dfs_abs2 \<equiv> do {
    REC (\<lambda>dfs v. do {
      if v = vt then
        RETURN True
      else do {
        FOREACHc
          ({u. (v,u)\<in>E}) (\<lambda>r. \<not>r) (\<lambda>u r. do {
          dfs u
        }) False
      }  
    }) v0
  }"

  (* Correctness lemma, easy direction. *)
  lemma
    assumes [simp]: "\<And>u. finite {v. (u,v)\<in>E}"
    shows "dfs_abs2 \<le> SPEC (\<lambda>r. r \<longrightarrow> (v0,vt)\<in>E\<^sup>*)"
    unfolding dfs_abs2_def
    apply (refine_vcg
      REC_rule[where pre="\<lambda>v. (v0,v)\<in>E\<^sup>*"]
      )
    apply auto []
    apply auto []
    
    apply (rename_tac dfs v)
    apply (rule_tac I="fe_inv v" in FOREACHc_rule)
    unfolding fe_inv_def
    apply auto []
    apply auto []
    apply auto []

    apply (rule order_trans)
    apply rprems
    apply auto []
    apply auto []
    thm FOREACHc_rule
    apply auto []
    apply auto []
    done


  (* Total correct variant. Assume wf T\<inverse> for correctness lemma *)

  definition "dfs_absT \<equiv> do {
    RECT (\<lambda>dfs v. do {
      if v = vt then
        RETURN True
      else do {
        FOREACHci (fe_inv v)
          ({u. (v,u)\<in>E}) (\<lambda>r. \<not>r) (\<lambda>u r. do {
          dfs u
        }) False
      }  
    }) v0
  }"

  lemma
    assumes WF: "wf (E\<inverse>)"
    assumes [simp]: "\<And>u. finite {v. (u,v)\<in>E}"
    shows "dfs_absT \<le> SPEC (\<lambda>r. r \<longrightarrow> (v0,vt)\<in>E\<^sup>*)"
    unfolding dfs_absT_def fe_inv_def
    apply (refine_vcg
      RECT_rule[where pre="\<lambda>v. (v0,v)\<in>E\<^sup>*", OF _ WF]
      )
    apply auto []
    apply auto []
    apply auto []
    apply auto []
    apply auto []

    apply (rule order_trans)
    apply rprems
    apply auto []
    apply auto []
    thm FOREACHc_rule
    apply auto []
    apply auto []
    apply auto []
    done




  end


  (* Show manual invar with rule_tac *)



  subsection \<open>While-Loop\<close>
  text \<open>An alternative to explore the nodes of a graph is a workset algorithm.
    It has the advantage that it allows multiple refinements of the workset, e.g.,
    to influence the search order.\<close>

  definition "abs_ws T v0 vt \<equiv> do {
    let s = ({v0},False);
    (* Iterate while workset is non-empty, and no path has been found.
      Note that we specify a partial correct loop here. *)
    s \<leftarrow> WHILE (\<lambda>(ws,r). \<not>r \<and> ws\<noteq>{}) (\<lambda>(ws,_). do {
      ASSERT (ws \<noteq> {});      (* Workaround quirk: Invariant not available for further refinement, so we assert it again *)
      v \<leftarrow> SPEC (\<lambda>x. x\<in>ws);  (* Pick element from workset *)
      let ws = ws - {v};     (* Remove it *)
      if v = vt then
        RETURN (ws,True)       (* If target node found, return true *)
      else do {
        let ws = ws \<union> T``{v};  (* Otherwise, add successors to workset *)
        RETURN (ws,False)
      }
    }) s;
    let (_,r) = s;    (* Extract the result *)
    RETURN r
  }"

  lemma abs_ws_correct: "abs_ws T v0 vt \<le> SPEC (\<lambda>r. r \<longrightarrow> (v0,vt)\<in>T\<^sup>*)"
    (* Invariant: I="\<lambda>(ws,r). ws \<subseteq> T\<^sup>*``{v0} \<and> (r \<longrightarrow> (v0,vt)\<in>T\<^sup>* )"*)
    unfolding abs_ws_def
    thm WHILET_rule
    apply (refine_vcg
      WHILE_rule[where I="\<lambda>(ws,r). ws \<subseteq> T\<^sup>*``{v0} \<and> (r \<longrightarrow> (v0,vt)\<in>T\<^sup>* )"]
    )
    apply auto
    by (metis Image_singleton_iff rtrancl_image_advance subsetCE)


  thm RETURN_refine bind_refine REC_refine 



end

